"""Define module exceptions."""


class SeventeenTrackError(Exception):
    """Define a base error."""

    pass


class InvalidTrackingNumberError(SeventeenTrackError):
    """Define an error for an invalid tracking number."""

    pass


class RequestError(SeventeenTrackError):
    """Define an error for HTTP request errors."""

    pass



"""Define profile-specific fixtures."""
import pytest


@pytest.fixture()
def authentication_failure_json():
    """Return a failed authentication response."""
    return {"Code": -6, "Message": "You haven't logged in for a long time."}


@pytest.fixture()
def authentication_success_json():
    """Return a failed authentication response."""
    return {
        "Json": {
            "FUserRole": 4,
            "FNickname": "John Doe",
            "FEmail": "john.doe@company.com",
            "FLanguage": "en",
            "FCountry": 100,
            "FPhoto": 10001,
            "gid": "1234567890987654321"
        },
        "Code": 0
    }


@pytest.fixture()
def packages_json():
    """Return package info from an account."""
    return {
        "pageInfo": {
            "Page": 1,
            "PerPage": 40,
            "TotalCount": 1
        },
        "Json": [{
            "FTrackInfoId": "1234567890987654321",
            "FTrackNo": "1234567890987654321",
            "FFirstCarrier": 0,
            "FFirstCarrierSource": 2,
            "FSecondCarrier": 0,
            "FSecondCarrierSource": 2,
            "FLastEvent": "{\"a\":\"2018-04-23 12:02\",\"b\":null,\"c\":\"\",\"d\":\"\",\"z\":\"Arrival at Destination Post\"}",
            "FIsArchived": False,
            "FRemark": "",
            "FTrackStateType": 0,
            "FCreateTime": "2018-05-04 20:22:13"
        }, {
            "FTrackInfoId": "1234567890987654321",
            "FTrackNo": "1234567890987654321",
            "FFirstCarrier": 0,
            "FFirstCarrierSource": 2,
            "FSecondCarrier": 0,
            "FSecondCarrierSource": 2,
            "FLastEvent": "",
            "FIsArchived": False,
            "FRemark": "",
            "FTrackStateType": 0,
            "FCreateTime": "2018-05-04 20:22:13"
        }]
    }


@pytest.fixture()
def summary_json():
    """Return a summary of packages in a profile."""
    return {
        "Json": {
            "utn": {
                "cnum": "40",
                "unum": "9",
                "inum": "1",
                "anum": 22
            },
            "eitem": [{
                "e": 10,
                "ec": 6
            }, {
                "e": 0,
                "ec": 2
            }, {
                "e": 20,
                "ec": 0
            }, {
                "e": 30,
                "ec": 0
            }, {
                "e": 35,
                "ec": 0
            }, {
                "e": 40,
                "ec": 0
            }, {
                "e": 50,
                "ec": 0
            }]
        },
        "Code": 0
    }



"""Define interaction with a user profile."""
import json
from typing import Coroutine, Callable, Union

from .package import PACKAGE_STATUS_MAP, Package

API_URL_BUYER = 'https://buyer.17track.net/orderapi/call'
API_URL_USER = 'https://user.17track.net/userapi/call'


class Profile:
    """Define a 17track.net profile manager."""

    def __init__(self, request: Callable[..., Coroutine]) -> None:
        """Initialize."""
        self._request = request
        self.account_id = None

    async def login(self, email: str, password: str) -> bool:
        """Login to the profile."""
        login_resp = await self._request(
            'post',
            API_URL_USER,
            json={
                'version': '1.0',
                'method': 'Signin',
                'param': {
                    'Email': email,
                    'Password': password,
                    'CaptchaCode': ''
                },
                'sourcetype': 0
            })

        if login_resp.get('Code') != 0:
            return False

        self.account_id = login_resp['Json']['gid']

        return True

    async def packages(
            self, package_state: Union[int, str] = '',
            show_archived: bool = False) -> list:
        """Get the list of packages associated with the account."""
        packages_resp = await self._request(
            'post',
            API_URL_BUYER,
            json={
                'version': '1.0',
                'method': 'GetTrackInfoList',
                'param': {
                    'IsArchived': show_archived,
                    'Item': '',
                    'Page': 1,
                    'PerPage': 40,
                    'PackageState': package_state,
                    'Sequence': '0'
                },
                'sourcetype': 0
            })

        packages = []
        for package in packages_resp.get('Json', []):
            last_event = package.get('FLastEvent')
            if last_event:
                event = json.loads(last_event)
            else:
                event = {}

            kwargs = {
                'destination_country': package.get('FSecondCountry', 0),
                'friendly_name': package.get('FRemark'),
                'info_text': event.get('z'),
                'location': event.get('c'),
                'origin_country': package.get('FFirstCountry', 0),
                'package_type': package.get('FTrackStateType', 0),
                'status': package.get('FPackageState', 0)
            }
            packages.append(Package(package['FTrackNo'], **kwargs))
        return packages

    async def summary(self, show_archived: bool = False) -> dict:
        """Get a quick summary of how many packages are in an account."""
        summary_resp = await self._request(
            'post',
            API_URL_BUYER,
            json={
                'version': '1.0',
                'method': 'GetIndexData',
                'param': {
                    'IsArchived': show_archived
                },
                'sourcetype': 0
            })

        results = {}
        for kind in summary_resp.get('Json', {}).get('eitem', []):
            results[PACKAGE_STATUS_MAP[kind['e']]] = kind['ec']
        return results



"""Define a simple structure for a package."""
import attr

COUNTRY_MAP = {
    0: 'Unknown',
    102: 'Afghanistan',
    103: 'Albania',
    104: 'Algeria',
    105: 'Andorra',
    106: 'Angola',
    108: 'Antarctica',
    110: 'Antigua and Barbuda',
    112: 'Argentina',
    113: 'Armenia',
    115: 'Australia',
    116: 'Austria',
    117: 'Azerbaijan',
    201: 'Bahamas',
    202: 'Bahrain',
    203: 'Bangladesh',
    204: 'Barbados',
    205: 'Belarus',
    206: 'Belgium',
    207: 'Belize',
    208: 'Benin',
    210: 'Bhutan',
    211: 'Bolivia',
    212: 'Bosnia and Herzegovina',
    213: 'Botswana',
    215: 'Brazil',
    216: 'Brunei',
    217: 'Bulgaria',
    218: 'Burkina Faso',
    219: 'Burundi',
    301: 'China',
    302: 'Cambodia',
    303: 'Cameroon',
    304: 'Canada',
    306: 'Cape Verde',
    308: 'Central African Republic',
    310: 'Chile',
    312: 'Ivory Coast',
    313: 'Colombia',
    314: 'Comoros',
    315: 'Congo-Brazzaville',
    316: 'Congo-Kinshasa',
    317: 'Cook Islands',
    318: 'Costa Rica',
    319: 'Croatia',
    320: 'Cuba',
    321: 'Cyprus',
    322: 'Czech Republic',
    323: 'Chad',
    401: 'Denmark',
    402: 'Djibouti',
    403: 'Dominica',
    404: 'Dominican Republic',
    501: 'Ecuador',
    502: 'Egypt',
    503: 'United Arab Emirates',
    504: 'Estonia',
    505: 'Ethiopia',
    506: 'Eritrea',
    507: 'Equatorial Guinea',
    508: 'East Timor',
    603: 'Fiji',
    604: 'Finland',
    605: 'France',
    701: 'Gabon',
    702: 'Gambia',
    703: 'Georgia',
    704: 'Germany',
    705: 'Ghana',
    707: 'Greece',
    709: 'Grenada',
    712: 'Guatemala',
    713: 'Guinea',
    714: 'Guyana',
    716: 'Guinea-Bissau',
    801: 'Hong Kong [CN]',
    802: 'Haiti',
    804: 'Honduras',
    805: 'Hungary',
    901: 'Iceland',
    902: 'India',
    903: 'Indonesia',
    904: 'Iran',
    905: 'Ireland',
    906: 'Israel',
    907: 'Italy',
    908: 'Iraq',
    1001: 'Jamaica',
    1002: 'Japan',
    1003: 'Jordan',
    1101: 'Kazakhstan',
    1102: 'Kenya',
    1103: 'United Kingdom',
    1104: 'Kiribati',
    1105: 'Korea, South',
    1106: 'Korea, North',
    1107: 'Kosovo',
    1108: 'Kuwait',
    1109: 'Kyrgyzstan',
    1201: 'Laos',
    1202: 'Latvia',
    1203: 'Lebanon',
    1204: 'Lesotho',
    1205: 'Liberia',
    1206: 'Libya',
    1207: 'Liechtenstein',
    1208: 'Lithuania',
    1209: 'Saint Lucia',
    1210: 'Luxembourg',
    1301: 'Macao [CN]',
    1302: 'Macedonia',
    1303: 'Madagascar',
    1304: 'Malawi',
    1305: 'Malaysia',
    1306: 'Maldives',
    1307: 'Mali',
    1308: 'Malta',
    1310: 'Marshall Islands',
    1312: 'Mauritania',
    1313: 'Mauritius',
    1314: 'Mexico',
    1315: 'Federated States of Micronesia',
    1316: 'Moldova',
    1317: 'Monaco',
    1318: 'Mongolia',
    1319: 'Montenegro',
    1321: 'Morocco',
    1322: 'Mozambique',
    1323: 'Myanmar',
    1401: 'Namibia',
    1402: 'Nauru',
    1403: 'Nepal',
    1404: 'Netherlands',
    1406: 'New Zealand',
    1407: 'Nicaragua',
    1408: 'Norway',
    1409: 'Niger',
    1410: 'Nigeria',
    1501: 'Oman',
    1601: 'Pakistan',
    1602: 'Palestine',
    1603: 'Panama',
    1604: 'Papua New Guinea',
    1605: 'Paraguay',
    1606: 'Peru',
    1607: 'Philippines',
    1608: 'Poland',
    1610: 'Portugal',
    1614: 'Palau',
    1701: 'Qatar',
    1802: 'Romania',
    1803: 'Russian Federation',
    1804: 'Rwanda',
    1902: 'Saint Vincent and the Grenadines',
    1903: 'El Salvador',
    1905: 'San Marino',
    1906: 'Sao Tome and Principe',
    1907: 'Saudi Arabia',
    1908: 'Senegal',
    1909: 'Serbia',
    1911: 'Seychelles',
    1912: 'Sierra Leone',
    1913: 'Singapore',
    1914: 'Slovakia',
    1915: 'Slovenia',
    1916: 'Solomon Islands',
    1917: 'South Africa',
    1918: 'Spain',
    1919: 'Sri Lanka',
    1920: 'Sudan',
    1921: 'Suriname',
    1923: 'Swaziland',
    1924: 'Sweden',
    1925: 'Switzerland',
    1926: 'Syrian Arab Republic',
    1927: 'Saint Kitts and Nevis',
    1928: 'Samoa',
    1929: 'Somalia',
    1930: 'Scotland',
    1932: 'South Ossetia',
    2001: 'Taiwan [CN]',
    2002: 'Tajikistan',
    2003: 'Tanzania',
    2004: 'Thailand',
    2005: 'Togo',
    2006: 'Tonga',
    2007: 'Trinidad and Tobago',
    2009: 'Tuvalu',
    2010: 'Tunisia',
    2011: 'Turkey',
    2012: 'Turkmenistan',
    2101: 'Uganda',
    2102: 'Ukraine',
    2103: 'Uzbekistan',
    2104: 'Uruguay',
    2105: 'United States',
    2202: 'Vanuatu',
    2203: 'Venezuela',
    2204: 'Vietnam',
    2205: 'Vatican City',
    2302: 'Western Sahara',
    2501: 'Yemen',
    2601: 'Zambia',
    2602: 'Zimbabwe',
    8901: 'Overseas Territory [ES]',
    9001: 'Overseas Territory [GB]',
    9002: 'Anguilla [GB]',
    9003: 'Ascension [GB]',
    9004: 'Bermuda [GB]',
    9005: 'Cayman Islands [GB]',
    9006: 'Gibraltar [GB]',
    9007: 'Guernsey [GB]',
    9008: 'Saint Helena [GB]',
    9101: 'Overseas Territory [FI]',
    9102: 'Ã…aland Islands [FI]',
    9201: 'Overseas Territory [NL]',
    9202: 'Antilles [NL]',
    9203: 'Aruba [NL]',
    9301: 'Overseas Territory [PT]',
    9401: 'Overseas Territory [NO]',
    9501: 'Overseas Territory [AU]',
    9502: 'Norfolk Island [AU]',
    9601: 'Overseas Territory [DK]',
    9602: 'Faroe Islands [DK]',
    9603: 'Greenland [DK]',
    9701: 'Overseas Territory [FR]',
    9702: 'New Caledonia [FR]',
    9801: 'Overseas Territory [US]',
    9901: 'Overseas Territory [NZ]'
}

PACKAGE_STATUS_MAP = {
    0: 'Not Found',
    10: 'In Transit',
    20: 'Expired',
    30: 'Ready to be Picked Up',
    35: 'Undelivered',
    40: 'Delivered',
    50: 'Returned'
}

PACKAGE_TYPE_MAP = {
    0: 'Unknown',
    1: 'Small Registered Package',
    2: 'Registered Parcel',
    3: 'EMS Package'
}


@attr.s(frozen=True)  # pylint: disable=too-few-public-methods
class Package:
    """Define a package object."""

    tracking_number = attr.ib()
    destination_country = attr.ib(default=0)
    friendly_name = attr.ib(default=None)
    info_text = attr.ib(default=None)
    location = attr.ib(default='')
    origin_country = attr.ib(default=0)
    package_type = attr.ib(default=0)
    status = attr.ib(default=0)
    tracking_info_language = attr.ib(default='Unknown')

    # pylint: disable=attribute-defined-outside-init
    def __attrs_post_init__(self):
        """Do some post-init processing."""
        object.__setattr__(
            self, 'destination_country', COUNTRY_MAP[self.destination_country])
        object.__setattr__(
            self, 'origin_country', COUNTRY_MAP[self.origin_country])
        object.__setattr__(
            self, 'package_type', PACKAGE_TYPE_MAP[self.package_type])
        object.__setattr__(
            self, 'status', PACKAGE_STATUS_MAP[self.status])



"""Define a 17track.net client."""
from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientError

from .errors import RequestError
from .profile import Profile
# from .track import Track


class Client:  # pylint: disable=too-few-public-methods
    """Define the client."""

    def __init__(self, websession: ClientSession) -> None:
        """Initialize."""
        self._websession = websession

        self.profile = Profile(self._request)

        # This is disabled until a workaround can be found:
        # self.track = Track(self._request)

    async def _request(
            self,
            method: str,
            url: str,
            *,
            headers: dict = None,
            params: dict = None,
            json: dict = None) -> dict:
        """Make a request against the RainMachine device."""
        if not headers:
            headers = {}

        try:
            async with self._websession.request(method, url, headers=headers,
                                                params=params,
                                                json=json) as resp:
                resp.raise_for_status()
                data = await resp.json(content_type=None)
                return data
        except ClientError as err:
            raise RequestError(
                'Error requesting data from {}: {}'.format(url, err))



"""Define interaction with an individual package."""
from typing import Callable, Coroutine

from .errors import InvalidTrackingNumberError
from .package import Package

API_URL_TRACK = 'https://t.17track.net/restapi/track'


class Track:  # pylint: disable=too-few-public-methods
    """Define a 17track.net package manager."""

    def __init__(self, request: Callable[..., Coroutine]) -> None:
        """Initialize."""
        self._request = request

    async def find(self, *tracking_numbers: str) -> list:
        """Get tracking info for one or more tracking numbers."""
        data = {'data': [{'num': num} for num in tracking_numbers]}
        tracking_resp = await self._request('post', API_URL_TRACK, json=data)

        print(tracking_resp)

        if not tracking_resp.get('dat'):
            raise InvalidTrackingNumberError('Invalid data')

        packages = []
        for info in tracking_resp['dat']:
            package_info = info.get('track', {})

            if not package_info:
                continue

            kwargs = {
                'destination_country': package_info.get('c'),
                'info_text': package_info.get('z0', {}).get('z'),
                'location': package_info.get('z0', {}).get('c'),
                'origin_country': package_info.get('b'),
                'package_type': package_info.get('d', 0),
                'status': package_info.get('e', 0),
                'tracking_info_language': package_info.get('ln1', 'Unknown')
            }
            packages.append(Package(info['no'], **kwargs))
        return packages
